
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <math.h>

void import_JPEG_file (const char* filename, unsigned char** image_chars,
		       int* image_height, int* image_width,
		       int* num_components);

int main (int argc, char *argv[]) {
  int m, n, c;
  int i, j, counter;
  int my_m, my_rank, num_procs;
  double val, avg, ex2, std_dv, n_avg, var;
  unsigned char *image_chars, *my_image_chars;
  char *input_jpeg_filename;
  int *scount, *displs;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);

  if(my_rank==0) {
    /* read from command line: input_jpeg_filename, output_jpeg_filename */
    if(argc < 2) {
      fprintf(stderr, "Give input_jpeg_filename as argument.\n");
      return -1;
    }

    input_jpeg_filename = argv[1];
    import_JPEG_file(input_jpeg_filename, &image_chars, &m, &n, &c);    
  }

  MPI_Bcast(&m, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  /* divide the m x n pixels evenly among the MPI processes */
  
  my_m = m/num_procs + (m%num_procs>my_rank);
  scount = malloc(num_procs*sizeof(int));
  displs = malloc(num_procs*sizeof(int));
  counter = 0;

  for(i = 0; i < num_procs; i++) {
    scount[i] = n*((m/num_procs) + (m%num_procs>i));
    displs[i] = counter;
    counter+=scount[i];
  }

  my_image_chars = malloc(my_m*n*sizeof(unsigned char));

  MPI_Scatterv(&image_chars[0], scount, displs, MPI_UNSIGNED_CHAR, &my_image_chars[0], my_m*n, MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
  
  /* Calculate average and standard deviation locally */
  avg = 0;
  std_dv = 0;
  
  for(i = 0; i < my_m; i++) {
    for(j = 0; j < n; j++) {
      val = (double)my_image_chars[i*n + j];
      avg += val/(m*n);
      ex2 += val*val/(m*n);   // Ex2
    }
  }

  MPI_Reduce(&avg, &n_avg, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce(&ex2, &var, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  
  if(my_rank==0) {  
    var = (m*n)/(m*n-1)*(var - n_avg*n_avg);
    std_dv = sqrt(var);
    printf("Average: %f, standard devation: %f\n", n_avg, std_dv);
  }

  free(my_image_chars);  
  free(scount);
  free(displs);
  if(my_rank==0)
    free(image_chars);
  
  MPI_Finalize();
  return 0;
}
